#include "stdafx.h"
#include <MersenneTwister.h>
#include <HMM.h>
#include "TaskMil.h"

void ODBG( const char * pch, COptionsMil * p ) { 
	char buf[300];
	if( p->m_oModel.m_aRates.Count()>1 ) 
		sprintf( buf, "%s [%f|%f --- %f|%f] [%f|%f --- %f|%f] LL=%f\n\n", 
				pch, 
				p->m_oModel.Rate(0)->m_adK0[0],			
				p->m_oModel.Rate(0)->m_adK1[0],			
				p->m_oModel.Rate(0)->m_adK0[1],			
				p->m_oModel.Rate(0)->m_adK1[1],			
				p->m_oModel.Rate(1)->m_adK0[0],			
				p->m_oModel.Rate(1)->m_adK1[0],			
				p->m_oModel.Rate(1)->m_adK0[1],			
				p->m_oModel.Rate(1)->m_adK1[1], 
				p->m_dLL );			
	else 
		sprintf( buf, "%s [%f|%f --- %f|%f] LL=%f\n\n", 
				pch, 
				p->m_oModel.Rate(0)->m_adK0[0],			
				p->m_oModel.Rate(0)->m_adK1[0],			
				p->m_oModel.Rate(0)->m_adK0[1],			
				p->m_oModel.Rate(0)->m_adK1[1],			
				p->m_dLL );			
	OutputDebugString( buf );
	}

COptionsMilStimulus * COptionsMil :: Stimulus( int iStim ) { 
	return (COptionsMilStimulus *) m_aStim[iStim];
	}
	
COptionsMilStimulus :: COptionsMilStimulus( const char * pch, int iChannel, double dValue, bool bValue ) { 
	m_cName.Set(pch);
	m_iChannel=iChannel;
	m_dValue=dValue;
	m_bValue=bValue;
	}
	
//----------------------------------------------------------------------
// COptionsMil registry load / save, members not persisted : 
// CModel m_oModel;		CModel * m_pOriginalModel;	CEventList m_oEventList;	
// double m_dSampling;	int m_iIterations;			double m_dLL;
//
#ifdef WIN32
bool COptionsMil :: RegistryLoad() { 
	CRegistryString o( "OPTIONS_MIL" );
	if( ! o.Load() ) 
		return false;

	int iVer = o.ExtractInt( );		
	m_iDataSelect = o.ExtractInt();
	m_iDataChannel = o.ExtractInt();
	m_dDeadTime = o.ExtractDouble();
	m_iChannels = o.ExtractInt();
	m_iMaxIter = o.ExtractInt();
	m_dConvLL = o.ExtractDouble();
	m_dConvGrad = o.ExtractDouble();
	m_cDataName = o.ExtractString();
	m_iMode = o.ExtractInt();
	m_bNumericDerivatives = o.ExtractBool();
	m_dK0Step = o.ExtractDouble();
	m_dK1Step = o.ExtractDouble();

	int iStims = o.ExtractInt( );		
	m_aStim.DeleteAll();
	for( int iStim=0; iStim<iStims; ++iStim ) {
		COptionsMilStimulus * pStim = new COptionsMilStimulus("", 0, 0.0, false );
		m_aStim.AddItem( pStim );
		pStim->m_cName=o.ExtractString();
		pStim->m_iChannel=o.ExtractInt();
		pStim->m_dValue=o.ExtractDouble();
		pStim->m_bValue=o.ExtractBool();
		}
				
	return true;
	}

bool COptionsMil :: RegistrySave() { 
	CRegistryString o( "OPTIONS_MIL" );
	
	int iVer=1;
	o.AppendInt( iVer );		
	o.AppendInt( m_iDataSelect );
	o.AppendInt( m_iDataChannel );
	o.AppendDouble( m_dDeadTime );
	o.AppendInt( m_iChannels );
	o.AppendInt( m_iMaxIter );
	o.AppendDouble( m_dConvLL );
	o.AppendDouble( m_dConvGrad );
	o.AppendString( m_cDataName.ConstBuffer() );
	o.AppendInt( m_iMode );
	o.AppendBool( m_bNumericDerivatives );
	o.AppendDouble( m_dK0Step );
	o.AppendDouble( m_dK1Step );
	
	int iStims = m_aStim.Count();
	o.AppendInt( iStims );		
	for( int iStim=0; iStim<iStims; ++iStim ) { 
		COptionsMilStimulus * pStim = Stimulus(iStim);
		o.AppendString( pStim->m_cName.ConstBuffer() );
		o.AppendInt( pStim->m_iChannel );
		o.AppendDouble( pStim->m_dValue );
		o.AppendBool( pStim->m_bValue );
		}

	return o.Save();
	}
	
#endif

COptionsMil :: COptionsMil() { 
	m_iDataChannel=0;
	m_dDeadTime=0.001;
	m_iDataSelect=0;
	m_iChannels=1;
	m_cDataName="";
	m_iMode=0;

	m_iMaxIter=1;
	m_dConvLL = 0.001;
	m_dConvGrad = 0.001;
	m_dSampling=0.0001;
	m_iIterations=0;
	m_dLL=0.0;
	
	m_bNumericDerivatives=true;
	m_dK0Step=1.0;
	m_dK1Step=1.0;
	}
	
COptionsMil :: ~COptionsMil() { 
	}
	
//====================================================== Calculations
// Mil Paper eqn 9 : part of the derivatives of the likelihood w.r.t Q matrix : 
double CTaskMil :: mil_gama( double dEig1, double dEig2, double dTDwell ) {
	if( dEig1==dEig2 ) 
		return dTDwell*exp(dEig1*dTDwell);
	return (exp(dEig1*dTDwell)-exp(dEig2*dTDwell))/(dEig1-dEig2);
	}


//-- The sum(...Aai * exp(LAMBDAai*t)...) portion of eq. (5) from the Mil Paper.
void CTaskMil :: mil_expQ( int iDim, double * w,  double ** padP, double ** padP1, double t, double** a) {
	for( int k=0; k<iDim; k++ ) {
		double dExp=exp(w[k]*t);
	    for( int i=0; i<iDim; i++ )
		    for( int j=0; j<iDim; j++ )
				a[i][j] += dExp * (padP[i][k]) * (padP1[k][j]);
		}
	}

void CTaskMil :: mil_backward( int iDwells, short * aiDwell, int * aiDuration, double dSampling ){
	Array1D<double> y( m_iMetaStates );
	Array1D<double> yy( m_iMetaStates );
	Array2D<double> a( m_iMetaStates, m_iMetaStates );

	int j, m, iClass, k2; //

	int iPrevClass = aiDwell[iDwells-1];
	int n = m_iMetaStates - m_oMetaModel.m_aiClassStateCount[iPrevClass];
	for( j=0; j<n; j++ ) 
		m_adBeta[2*iDwells][j]=m_adScale[2*iDwells];		//s
	
	CMilModelData * pData;
	for( int k=iDwells-1; k>=0 && ! m_bCancelled; k-- ) {
		pData = &(m_pData[ m_pOptionsMil->m_oEventList.m_aiStimIndex==NULL 
						? 0 : m_pOptionsMil->m_oEventList.m_aiStimIndex[k] ]);

		k2 = 2*k;
		iClass = aiDwell[k];
		m  = m_oMetaModel.m_aiClassStateCount[iClass];
		
		//- Multiply extract from m_aQQE by m_adBeta[k*2+2] into y[]
		// Note n =  #not in class in first iteration, # in class for rest.
		matextractmult( m, n, 
					m_oMetaModel.m_aiClassStateIndex[iClass], 
					(m_oMetaModel.m_aiClassStateIndex[iPrevClass])+( k==iDwells-1?m_iMetaStates:0),
					(double **)pData->m_aQQe, (double *)m_adBeta[k2+2], (double *)y );
		
		for (j=0; j<m; j++) { 
			if ( (y[j] > 1.0e120) && (m_adScale[k2+1] > 1.0) ) // Chris attempt to avoid overflow 11-19-02
				m_adBeta[k2+1][j] = 1.0e120;
			else if ( (y[j] < 1.0e-120) && (m_adScale[k2+1] < 1.0) )
				m_adBeta[k2+1][j] = 1.0e-120;						// Chris attempt to avoid overflow 5.21.02 rev 11-19-02
			else
				m_adBeta[k2+1][j]=y[j]*m_adScale[k2+1];               // float overflow
			}
		
		//-- Mil paper eq.(5)
		Array2D<double> aTemp( m, m, 0.0 );
		mil_expQ( m, pData->m_aW[iClass], pData->m_adP[iClass], pData->m_adP1[iClass], dSampling*aiDuration[k]-m_pOptionsMil->m_dDeadTime, aTemp );
		yy = matmult( m, m,(double **) aTemp, (double *)m_adBeta[k2+1] );
		
		for (j=0; j<m; j++) {
			if ( (yy[j] > 1.0e120) && (m_adScale[k2] > 1.0) ) // Chris attempt to avoid overflow 11-19-02
				m_adBeta[k2][j] = 1.0e120;
			else if ( (yy[j] < 1.0e-120) && (m_adScale[k2] < 1.0) )
				m_adBeta[k2][j] =1.0e-120;						// Chris attempt to avoid overflow 5.21.02 rev 11-19-02
			else
				m_adBeta[k2][j]=yy[j]*m_adScale[k2];                   // float overflow
			}
		
		iPrevClass = iClass;
		n  = m;
		}
	}

//-- dLL/dQQE
//-- TODO - is it correct for Qaa and Qab to process last dwell? Perhaps,
// since it is not the same info as the other dwells.  Seems to calc the 
// probability of exiting into any other state, but last dwell might be 
// a partial dwell
//-- Note : could pre-calc the b arrays (limited usefulness?)
void CTaskMil :: mil_dlogLL_qe( ) {
	int iDwells=m_pOptionsMil->m_oEventList.m_iEvents;
	short * aiDwell=m_pOptionsMil->m_oEventList.m_aiClass;
	int * aiDuration=m_pOptionsMil->m_oEventList.m_aiDuration;
	unsigned short * aiStimIndex=m_pOptionsMil->m_oEventList.m_aiStimIndex;
	double dSampling=m_pOptionsMil->m_dSampling;
	int iClsA, iClsB;
	int	 i,j,ii,jj,m,iDwell;
	double tmp;
	CMilModelData * pData;

	//--
	for( i=0; i==0 || i<m_pOptionsMil->m_oEventList.m_iStimCount; i++)
		m_pData[i].m_df_qqe = 0.0;

	//-- 
	for( iClsA=0; iClsA<m_iMetaClasses && ! m_bCancelled; iClsA++) {
		m=m_oMetaModel.m_aiClassStateCount[iClsA];
		Array2D<double> Ai(m,m);
		Array2D<double> aby(m,m);
		Array2D<double> Aj(m,m);
		Array2D<double> Temp(m,m);

		//-- Part 1 - Qaa - Eq. (8) from paper 
		//TODO - once this all works, precalc of Ai (a.k.a. Aj) 
		for( ii=0; ii<m; ii++ ) {
			for( jj=0; jj<m; jj++) {
				for( iDwell=0; iDwell<iDwells; iDwell++ ) {	// [ sum alpha.ii-1 * beta.ii+1Transpose * gamma.a.i.j(t) ]
					if ( aiDwell[iDwell]==iClsA) {
						//pData = &m_pData[ aiStimIndex[ iDwell ] ];
						pData = &(m_pData[ aiStimIndex==NULL ? 0 : aiStimIndex[iDwell] ]);
						
						//-- Aa,i
						for( i=0; i<m; i++ ) 
							for( j=0; j<m; j++ ) 
								Ai[i][j]=pData->m_adP[iClsA][j][ii]*pData->m_adP1[iClsA][ii][i]; 
						
						//-- S[ii=1..jj]( a *  * y(t) ) 
						tmp=mil_gama( pData->m_aW[iClsA][ii], pData->m_aW[iClsA][jj], dSampling*aiDuration[iDwell]-m_pOptionsMil->m_dDeadTime );
						for( i=0; i<m; i++ ) 
							for( j=0; j<m; j++ ) 
								aby[i][j] = m_adAlpha[2*iDwell][i]*m_adBeta[2*iDwell+1][j]*tmp;
								
						//-- Aa,j Transpose  Aj(state j)
						for (i=0; i<m; i++) 
							for (j=0; j<m; j++) 
								Aj[i][j] = pData->m_adP[iClsA][j][jj]*pData->m_adP1[iClsA][jj][i];           
						
						//-- Mil Eqn. (8) m_df_qqe += Ai * aby * Aj
						MxMxMIntoMap( m, (double **)Ai, (double**)aby, (double**)Aj, 
								m_oMetaModel.m_aiClassStateIndex[iClsA], 
								m_oMetaModel.m_aiClassStateIndex[iClsA], 
								(double **)pData->m_df_qqe );
						}
					}
				}
			}
		}
			

	//-- part 2 - Qab a!=b  [ From MIL paper eqn (7)   dF/dQQEab ]
	// Aai and exp(EIGi Tk) are stored in extra elements of alpha and beta  TODO - alpha and beta should be split to two arrays each ? 
	// dF/dQQE [states in class of dwell(iDwell)][states in class of dwell(iDwell+1)] += alpha[dwell-1] X beta[dwell+1]
	for( iDwell=0; iDwell+1<iDwells && ! m_bCancelled; iDwell++) 
		if( (iClsB=aiDwell[iDwell+1]) != (iClsA=aiDwell[iDwell]) ) {
			//pData = &m_pData[ aiStimIndex[ iDwell ] ];
			pData = &(m_pData[ aiStimIndex==NULL ? 0 : aiStimIndex[iDwell] ]);
			vecmultintomap( m_oMetaModel.m_aiClassStateCount[iClsA], 
							m_oMetaModel.m_aiClassStateCount[iClsB], 
							m_oMetaModel.m_aiClassStateIndex[iClsA], 
							m_oMetaModel.m_aiClassStateIndex[iClsB], 
							(double *) m_adAlpha[2*iDwell+1], 
							(double *) m_adBeta[2*iDwell+2],
							(double **)pData->m_df_qqe, true );
			}

	// Last dwell exit to any class other than this.
	// Note : for partial dwell selection, skip this ? 
	iClsA=aiDwell[(iDwells-1)];
	for( iClsB=0; iClsB<m_iMetaClasses && ! m_bCancelled; iClsB++ )
		if( iClsB!=iClsA ){		
			//pData = &m_pData[ aiStimIndex[ iDwells-1 ] ];
			pData = &(m_pData[ aiStimIndex==NULL ? 0 : aiStimIndex[iDwell] ]);
			vecmultintomap( m_oMetaModel.m_aiClassStateCount[iClsA], 
							m_oMetaModel.m_aiClassStateCount[iClsB], 
							m_oMetaModel.m_aiClassStateIndex[iClsA], 
							m_oMetaModel.m_aiClassStateIndex[iClsB], 
							(double *) m_adAlpha[2*(iDwells-1)+1], 
							(double *) m_adBeta[2*(iDwells-1)+2],
							(double **) pData->m_df_qqe, true );
			}
	}

//-----------------------------------------------------------------
CTaskMil :: CTaskMil() {
	m_pOptionsMil=NULL;
	m_pData=NULL;
	}
	
CTaskMil :: ~CTaskMil() {
	// TODO 
	//-- create results window 		
	//pTaskMil->m_hwndDataView = OpenDataView( pConfig->m_cFileName.ConstBuffer() );
	//PostMessage( pTaskMil->m_hwndDataView, WM_DATA_CREATENEW, (WPARAM)pConfig, 0 );	

	if( m_pOptionsMil!=NULL ) {
		delete m_pOptionsMil;
		}

	if( m_pData!=NULL ) {
		delete [] m_pData;
		}
		
	}
			
	
//----------------------------------------------------------------- MIL 
//-- # of terms required in power series expansion 
int mil_mexp_terms( double ** q, int n, double td ) {	   
	const double EPS= 1.0E-50;

	double dSum=0.0;
	for( int i=0; i<n; i++ ) 
		for( int j=0; j<n; j++ ) 
			dSum += fabs(q[i][j]);
	dSum *= td;

	double r=1.0;
	for( int i=0; r/(1.0-r/(2.0+i))>=EPS; i++) 
		r *= dSum/(i+1);

	return max(50,i);
	}

//-- matrix exponential as power series expansion 
// From Mil paper :
// (19) Raa = sum( U{k=0..n} ) 
// (20) dRaa/dQaij = sum( V{k=0..n} ) 
// (21) U{k+1} = t/(k+1) U{k}Qaa
// (22) V{k+1} = t/(k+1) ( Qaa V{k} + Eij U{k}
// Eij is single element [i][j]=1  
// Note : this implementation uses dq_x instead of E.
void mil_mexp( Array2D<double> & q, Array3D<double> * pdq_x, int n, int iPaths, double td, 
			   Array2D<double> & a, Array3D<double> * pda_x, int ie, bool bDerivatives) {
	Array2D<double> u[2];						// u from eqn.(21) 
	u[0]=Array2D<double>( n, n, 0.0 );
	u[1]=Array2D<double>( n, n, 0.0 );
	u[0].SetDiagonal( ie==0 ? 1.0 : td );		// set diagonal of u

	Array3D<double> v[2];						// v from eqn.(22)
	v[0]=Array3D<double>( iPaths, n, n, 0.0);
	v[1]=Array3D<double>( iPaths, n, n, 0.0);
	
	a = 0.0;										// Initialize the sums to 0's
	if( bDerivatives ) 
		*pda_x = 0.0;
		
	int iIterations = mil_mexp_terms(q,n,td);		// terms required for power series expansion of exp.
	for( int iIter=0; iIter<iIterations; ++iIter ){ 
		int iSrc = ( iIter%2==0 ? 0 : 1 );			//-- whether going v[0]->v[1] or v[1]->v[0].
		
		if( bDerivatives ) {
			//-- v1[i] = q * v[i] + dq_x[path] * u;  v1 *= td/(iIter+1+ie)
			*pda_x += v[iSrc];							//-- da_x = da_x + v
			for( int iPath=0; iPath<iPaths; iPath++ ) {	
				matmult( n, n, n, (double **)q,  (double **)v[iSrc][iPath], (double **) v[1-iSrc][iPath] );	// v[i] = q * v[i-1]
				matmult( n, n, n, (double **)(*pdq_x)[iPath],  (double **)u[iSrc], (double **) v[1-iSrc][iPath], true );		// v[i] += dq_x[path] * u 
				}										
			v[1-iSrc].MultiplyAdd( td/(iIter+1+ie), 0.0 );		
			}

		//-- u = u * q * td/(iIter+1+ie)
		a += u[iSrc];								//-- a = a + u
		matmult( n, n, n, (double **)u[iSrc], (double **)q, (double **)u[1-iSrc] );	
		u[1-iSrc].MultiplyAdd( td/(iIter+1+ie), 0.0 );
		}
	}

//-- Qab Qbc calc's
void CTaskMil :: qabc(	int na, int nb, int nc, 
			Array1D<int> & ida, Array1D<int> & idb, Array1D<int> & idc, 
			Array2D<double> & qab, Array2D<double> & qbb, Array2D<double> & qbc, 
			double td, 
			Array2D<int> & path, 
			int iPaths, 
			Array2D<double> & qac, 
			Array3D<double> * pdqac_q ) {

	//-- set totals to 0 then check if nothing to do			
	qac = 0.0;
	if( m_bAnalyticalDerivative )
		*pdqac_q = 0.0;
	if( nb==0 )
		return;
		
	int	i,j;
	int iMax=max(max(na,nb),nc)+1;
	Array2D<double> a( iMax, iMax );
	Array2D<double> temp( iMax, iMax );

	if( m_bAnalyticalDerivative ) {
		//-- d(Qac)/d(Q) : 
		Array3D<double> dqbb_q( iPaths, iMax, iMax, 0.0 );
		Array3D<double> da_q( iPaths, iMax, iMax );
		Array2D<double> vv( iMax, iMax );
		//-- set nonzero elements of dqbb_q where [i][j] correspond to the same point 
		// or 2 path connected points in original model
		for( int iPath=0; iPath<iPaths; iPath++ )  
			for( i=0; i<nb; i++ ) 
				for( j=0; j<nb; j++ ) {
					if( idb[i]==path[iPath][0] ) {
						if( idb[j]==path[iPath][1] ) // [i][j] correspond to a path 
							dqbb_q[iPath][i][j] = 1;
						else if( idb[j]==idb[i] )    // [i][j] correspond to the same point
							dqbb_q[iPath][i][j] = -1;
						}
					}    
	
		mil_mexp( qbb, &dqbb_q, nb, iPaths, td, a, &da_q, 1, true );
		
		//-- 
		for( int iPath=0; iPath<iPaths; iPath++ ) {
			//-- vmap_ab = map of elements which correspond to path[iPath]
			Array2D<double> vmap_ab( na, nb, 0.0 );
			for( i=0; i<na; i++ )
				if( ida[i]==path[iPath][0] ) 
					for( j=0; j<nb; j++ ) 
						if( idb[j]==path[iPath][1] ) 
							vmap_ab[i][j] = 1.0;
					
			//-- dqac_q[iPath] += vmap_ab * a * qbc
			matmult( na, nb, nb,  (double **)vmap_ab,  (double **)a, (double **)temp );
			matmult( na, nb, nc,  (double **)temp,  (double **)qbc, (double **)(*pdqac_q)[iPath], true );

			//--  dqac_q[iPath] += Qab * da_Q[iPath] * Qbc
			matmult( na, nb, nb,  (double **)qab,  (double **)da_q[iPath], (double **)temp );
			matmult( na, nb, nc,  (double **)temp,  (double **)qbc, (double **)(*pdqac_q)[iPath], true );

			//----- temp array only nonzero in a few places, perhaps a better method ? todo 
			Array2D<double> vmap_bc( nb, nc, 0.0 );
			for (i=0; i<nb; i++) 
				if( idb[i]==path[iPath][0] ) 
				for (j=0; j<nc; j++) 
						if( idc[j]==path[iPath][1] ) 
							vmap_bc[i][j] = 1;			// copy column i of a to col j of vv

			// qab*(a*v)
			//-- dqac_q[iPath][0..na][0..nb] += ( qab*(a*vmap_bc) )[0..na][0..nb]
			matmult( nb, nb, nc, (double **)a, (double **)vmap_bc, (double **) vv );
			matmult( na, nb, nc, (double **)qab, (double **)vv, (double **) (*pdqac_q)[iPath], true );
			}
		}
	else 
		mil_mexp( qbb, NULL, nb, iPaths, td, a, NULL, 1, false );

	//-- Qac = Qab * R * Qbc	[ last term of eqn. (16) ] 
	matmult( na, nb, nb, (double **)qab, (double **)a, (double **)temp );			
	matmult( na, nb, nc, (double **)temp, (double **)qbc, (double **)qac );
	}


//-- missed events correction from q to qe ( and qQe/dQ )
void CTaskMil :: qtoqe( int iStates, 
			int iPaths, 
			Array2D<int> & path, 
			Array1D<int> & aiClass, 
			Array1D<int> & aiStateClass, 
			int iClasses,
			double td,
			CMilModelData * pData,  
			Array2D<double> & q ){

	int i, j, k, iClass, jClass, si, sj, na, nb, nc; 
	double tmp;
	
	Array2D<double> qab( iStates, iStates );
	Array2D<double> qbb( iStates, iStates );
	Array2D<double> e( iStates, iStates );

	//-- These are only used with m_bAnalyticalDerivative, but defined here for efficiency
	// rather than recreating in loop.  Create very small if not needed.
	int iSt = m_bAnalyticalDerivative ? iStates : 1;
	int iPth = m_bAnalyticalDerivative ? iPaths : 1;
	Array3D<double> de_q( iPth, iSt, iSt );
	Array3D<double> dqbb_q( iPth, iSt, iSt );
	Array3D<double> dqab_q( iPth, iSt, iSt );
   
	//-- Correct diagonal blocks 
	for( iClass=0; iClass<iClasses; iClass++) {
		//-- Construct map of states in class and not in class
		Array1D<int> iMapA = matextractmap<int>( iStates, aiStateClass, aiClass[iClass], true );
		Array1D<int> iMapC = matextractmap<int>( iStates, aiStateClass, aiClass[iClass], false );
		na=iMapA.dim();
		nc=iMapC.dim();

		qabc( na, nc, na, iMapA, iMapC, iMapA,
			matextract( na, nc, iMapA, iMapC, (double **) q ),		// qac
			matextract( nc, nc, iMapC, iMapC, (double **) q ),		// qcc
			matextract( nc, na, iMapC, iMapA, (double **) q ),		// qca
			td, path, iPaths, e, m_bAnalyticalDerivative ? &de_q : NULL );

		for( i=0; i<na; i++) {
			si=iMapA[i];
			for( j=0; j<na; j++) {
				sj=iMapA[j];
				pData->m_aQQe[si][sj]=q[si][sj]+e[i][j];
				if( m_bAnalyticalDerivative ) { 
					for( k=0; k<iPaths; k++) {
						if( si==path[k][0] && sj==path[k][1] ) 
							tmp = 1.0;
						else if( si==path[k][0] && sj==si )    
							tmp =-1.0;
						else 				
							tmp = 0.0;
						pData->m_dQQe_qq[k][si][sj]=tmp+de_q[k][i][j];
						}
					}
				}
			}
		}
	
	
	// correct off-diagonal blocks 
	Array1D<int> ida( iStates );
	Array1D<int> idb( iStates );
	Array1D<int> idc( iStates );
	for( iClass=0; iClass<iClasses; iClass++) {
        for( jClass=0; jClass<iClasses; jClass++) {
			if( iClass!=jClass ) {
				//-- Construct maps : ida[na]=matches to iClass; 
				// idb[nb]=matches to jClass; idc[nc]=the rest.
				na=nb=nc=0;
				for( i=0; i<iStates; ++i ) {
					if( aiStateClass[i]==aiClass[iClass] ) 
						ida[na++]=i;
					else if( aiStateClass[i]==aiClass[jClass] ) 
						idb[nb++]=i;
					else 
						idc[nc++]=i;
					}
					
				qabc( na, nc, nb, ida, idc, idb, 
					matextract( na, nc, ida, idc, (double **)q ),		// qac
					matextract( nc, nc, idc, idc, (double **)q ),		// qcc
					matextract( nc, nb, idc, idb, (double **)q ),		// qcb
					td, path, iPaths, e, m_bAnalyticalDerivative ? &de_q : NULL );

				for( i=0; i<na; i++) 
					for( j=0; j<nb; j++)
						qab[i][j]=q[ida[i]][idb[j]]+e[i][j];
					
				if( m_bAnalyticalDerivative ) {
					for( i=0; i<na; i++) {
						si=ida[i];
						for( j=0; j<nb; j++) {
							sj=idb[j];
							for( k=0; k<iPaths; k++) {
								if( si==path[k][0] && sj==path[k][1] ) 
									tmp= 1.0;
								else 				
									tmp= 0.0;
								dqab_q[k][i][j]=tmp+de_q[k][i][j];
								}
							}
						}

					for( i=0; i<nb; i++) 
						for( j=0; j<nb; j++) {
							si=idb[i];
							sj=idb[j];
							qbb[i][j]=q[si][sj];
							if( m_bAnalyticalDerivative ) { 
								for( k=0; k<iPaths; k++) { 
									if( si==path[k][0] && sj==path[k][1] ) 
										tmp= 1;
									else if( si==path[k][0] && sj==si )    
										tmp=-1;
									else 				
										tmp= 0;
									dqbb_q[k][i][j]=tmp;
									}
								}    
							}    

					mil_mexp( qbb, &dqbb_q, nb, iPaths, td, e, &de_q, 0, true );
					//-- mapped rows/cols of m_aQQe = qab[na,nb] * e[nb,nb]
					matmultintomap<double>( na, nb, nb, ida, idb, qab, e, pData->m_aQQe );

					for( k=0; k<iPaths; k++ ) {
						//-- mapped rows/cols of m_dQQe_qq = dqab_q[k] * e 
						matmultintomap<double>( na, nb, nb, ida, idb, dqab_q[k], e, pData->m_dQQe_qq[k] );
						//-- mapped rows/cols of m_dQQe_qq = qab * de_q[k]
						matmultintomap<double>( na, nb, nb, ida, idb, qab, de_q[k], pData->m_dQQe_qq[k], true );
						}
					}
				else {  
					for( i=0; i<nb; i++) 
						for( j=0; j<nb; j++)
							qbb[i][j]=q[idb[i]][idb[j]];
					mil_mexp( qbb, NULL, nb, iPaths, td, e, NULL, 0, false );
					//-- mapped rows/cols of m_aQQe = qab[na,nb] * e[nb,nb]
					matmultintomap<double>( na, nb, nb, ida, idb, qab, e, pData->m_aQQe );
					}
				}
			}
		}
	}	        
	    
//-----------------------------------------------------------------
/*
pModel:	
	->	X			\
	->	mtx, vct	/->	Z, dx_z
	->	dq_x
	->	oMetaModel:	
		->	m_adInitPr		\
		->	m_aiPaths		|
		->	aiStateClasses	|-(qtoqe)->m_aQQE, m_dQQE_q:	
		->	ClassStateIndex	|				->	adQQE_k:	
		->	QQ,dQQ_Q		/					-> m_aW, m_adP, m_adP1
*/						

void CTaskMil :: ModelSetup( CModel * pModel, CMilModelData * pData ) { 
	
	if( m_bAnalyticalDerivative ) { 
		//-- Create X
		m_iOrigRates = pModel->m_aRates.Count();
		m_iOrigPaths = 2 * m_iOrigRates;
		m_iXSize = 2 * m_iOrigPaths;
		pData->m_X = Array1D<double>( m_iXSize, 0.0 );
		for( int iP=0; iP<m_iOrigRates; ++iP ) {
			CModelRate * pRate = pModel->Rate( iP );
			pData->m_X[iP*4+0] = log( pRate->RateValueK0( true ) );
			pData->m_X[iP*4+1] = pRate->RateValueK1( true );
			pData->m_X[iP*4+2] = log( pRate->RateValueK0( false ) );
			pData->m_X[iP*4+3] = pRate->RateValueK1( false );
			}

		//-- X to Z			
		int iConstraintRows;
		pModel->BuildConstraintsMtxVct( pData->m_Mtx, pData->m_Vct, &iConstraintRows );
		pModel->FreePar( iConstraintRows, pData->m_X, pData->m_Mtx, pData->m_Vct, pData->m_Z );
		m_iZSize=pData->m_Z.dim();

		//-- dx_z
		pData->m_dx_z = Array2D<double>( m_iXSize, m_iZSize, 0.0 );
		matcopy( pData->m_dx_z.dim1(), pData->m_dx_z.dim2(), (double **)pData->m_Mtx, (double **)pData->m_dx_z );
		}
	
	//-- meta model 
	m_oMetaModel.Setup( &m_pOptionsMil->m_oModel, m_pOptionsMil->m_iChannels, false );

	//-- from Metamodel.m_aRates -> m_aiPaths
	m_iMetaPaths = m_oMetaModel.m_aPaths.Count()*2;		
	m_aiPaths = Array2D<int>( m_iMetaPaths, 2 );
	m_oMetaModel.FillPathArray( m_aiPaths, NULL, 0.0 );

	//-- from metamodel->m_aStates -> aiStateClasses, adInitPr
	m_iMetaStates = m_oMetaModel.m_aStates.Count();		
	m_aiStateClasses = Array1D<int>( m_iMetaStates );
	m_adInitPr = Array1D<double>( m_iMetaStates, 0.0 );
			
	for( int iState=0; iState<m_iMetaStates; ++iState ) {
		m_aiStateClasses[iState]=m_oMetaModel.State(iState)->m_iClass;	
		m_adInitPr[iState]=m_oMetaModel.State(iState)->m_dStartProb;
		}

	m_iMetaClasses = m_oMetaModel.UsedClassCount( &m_aiClasses );		
	m_oMetaModel.SetupClassStateIndex( m_iMetaClasses, m_iMetaStates );

	pData->m_aQQe = Array2D<double>( m_iMetaStates, m_iMetaStates );
	
	pData->m_aW = Array2D<double>( m_iMetaClasses, m_iMetaStates, 0.0 );						
	pData->m_adP = Array3D<double>( m_iMetaClasses, m_iMetaStates, m_iMetaStates, 0.0 );
	pData->m_adP1 = Array3D<double>( m_iMetaClasses, m_iMetaStates, m_iMetaStates, 0.0 );
	
	//-- Mil gradients setup 
	if( m_bAnalyticalDerivative ) { 
		pData->m_dq_x = Array2D<double>( m_iOrigPaths, m_iXSize );
		pData->m_dqq_q = Array2D<double>( m_iMetaPaths, m_iOrigRates*2 );
		pData->m_dQQe_qq = Array3D<double>( m_iMetaPaths, m_iMetaStates, m_iMetaStates );
		pData->m_df_qqe = Array2D<double>( m_iMetaStates, m_iMetaStates );
		pData->m_df_qq = Array1D<double>( m_iMetaPaths );
		pData->m_df_q = Array1D<double>( m_iOrigPaths );
		pData->m_df_z = Array1D<double>( m_iZSize );
		pData->m_df_x = Array1D<double>( m_iXSize );
		}
	}

void CTaskMil :: ModelMultipleSetup( COptionsMil * pOpt ) { 
	//-- For each stimulus set ... ( at least one ! ) 
	for( int iSet=0; iSet==0 || iSet<pOpt->m_oEventList.m_iStimCount; ++iSet ) {
		ModelSetup( &pOpt->m_oModel, &m_pData[iSet] );
		ModelCalculate( &pOpt->m_oModel, pOpt->m_dDeadTime, &m_pData[iSet] );
		}
	}
	
void CTaskMil :: ModelCalculate( CModel * pModel, double dDeadTime, CMilModelData * pData ) { 
	if( m_bAnalyticalDerivative ) { 
		//-- m_dq_x ( a.k.a. ) dQ/dX
		// todo - pre-calc nQK0 and nQK1 for all models & levels ?
		double nQK0, nQK1;
		pData->m_dq_x = 0.0;
		for( int iPath=0; iPath<m_iOrigRates; ++iPath ) { 
			CModelRate * pRate = pModel->Rate( iPath );
			
			nQK0 = pRate->RateValue( true, NULL, 0.0 );		//todo add stimulus
			nQK1 = pRate->RateValueK1( true, NULL, 0.0 );	//todo add stimulus
			pData->m_dq_x[iPath*2][4*iPath] = nQK0;
			pData->m_dq_x[iPath*2][4*iPath+1] = nQK1 * nQK0;
	
			nQK0 = pRate->RateValue( false, NULL, 0.0 );	//todo add stimulus
			nQK1 = pRate->RateValueK1( false, NULL, 0.0 );	//todo add stimulus
			pData->m_dq_x[iPath*2+1][4*iPath+2] = nQK0;
			pData->m_dq_x[iPath*2+1][4*iPath+3] = nQK1 * nQK0;
			}
		}
		
	//-- Metamodel QQ and dQQ/dq		
	Array2D<double> aQ = m_oMetaModel.QQ( NULL/*todo add stimulus*/, 0.0, 
						m_bAnalyticalDerivative ? &(pData->m_dqq_q) : NULL );	

	//-- QQe, dQQe/dQ
	qtoqe( m_iMetaStates, 
			m_iMetaPaths, 
			m_aiPaths, 
			m_aiClasses, 
			m_aiStateClasses, 
			m_iMetaClasses, 
			dDeadTime, 
			pData, 
			aQ );

	//-- eigenvalues / vectors - spectral decomposition
	for( int iClass=0; iClass<m_iMetaClasses; iClass++) {
		int iSize = m_oMetaModel.m_aiClassStateCount[iClass];
		Array2D<double> adQQE_k = matextract( iSize, iSize, m_oMetaModel.m_aiClassStateIndex[iClass], m_oMetaModel.m_aiClassStateIndex[iClass], (double **) pData->m_aQQe );
		//if ( _isnan(a[i][j]) || a[i][j] == DBL_MAX || a[i][j] > 10e20) throw a[i][j];
		if( ! Spectrum( adQQE_k, (double *)pData->m_aW[iClass], (double **)pData->m_adP[iClass], (double **)pData->m_adP1[iClass] ) ) {
			//TODO - error 
			}
		}
		
	}
	
void CTaskMil :: EvaluateSetup( int iEvents ) { 
	m_adAlpha = Array2D<double>( iEvents*2+1, m_iMetaStates, 0.0 );
	m_adScale = Array1D<double>( iEvents*2+1, 0.0 );
	if( m_bAnalyticalDerivative ) 
		m_adBeta = Array2D<double>( iEvents*2+1, m_iMetaStates, 0.0 );

	/*	
	unsigned short * aiStimIndex=m_pOptionsMil->m_oEventList.m_aiStimIndex;
	int * aiDuration=m_pOptionsMil->m_oEventList.m_aiDuration;
	int i, iDwell, iDwells = m_pOptionsMil->m_oEventList.m_iEvents;

	//-- set magnitudes - sum of dwell times per stimulii set.
	if( m_pOptionsMil->m_oEventList.m_iStimCount>1 ) { 
		for( i=0; i<m_pOptionsMil->m_oEventList.m_iStimCount; i++)
			m_pData[i].m_dMagnitude=0.0;
		for( iDwell=0; iDwell<iDwells; iDwell++ )
			m_pData[ aiStimIndex[iDwell] ].m_dMagnitude += aiDuration[iDwell];
		for( i=0; i<m_pOptionsMil->m_oEventList.m_iStimCount; i++)
			m_dMagnitudeSum += m_pData[i].m_dMagnitude;
		}
	else 
		m_dMagnitudeSum =m_pData[0].m_dMagnitude=1.0;
	*/
		
	}
	
double CTaskMil :: EvaluateLL( ) {		// the 'forward' part of forward-backward.
	int iDwells=m_pOptionsMil->m_oEventList.m_iEvents;
	short * aiDwell=m_pOptionsMil->m_oEventList.m_aiClass;
	int * aiDuration=m_pOptionsMil->m_oEventList.m_aiDuration;
	unsigned short * aiStimIndex=m_pOptionsMil->m_oEventList.m_aiStimIndex;
	double dSampling=m_pOptionsMil->m_dSampling;

	Array2D<double> a( m_iMetaStates, m_iMetaStates );

	int i,j,k,m,n,iClass,iNextClass,k2;
	double tmp;
	double dScale;

	Array1D<double> aP0( m_iMetaStates );

	for( int jj=0; jj<2*iDwells+1; ++jj ) 
		m_adAlpha[jj][0]=-1.0;
		
	iClass = aiDwell[0];
	m  = m_oMetaModel.m_aiClassStateCount[iClass];

	//-- starting pr of the states in the first dwell 
	tmp=0.0;
	for( i=0; i<m; i++ )
		tmp += m_adInitPr[m_oMetaModel.m_aiClassStateIndex[iClass][i]];
	if( tmp==0.0 ) 
		aP0 = 1.0/m;
	else 
		for( i=0; i<m; i++ )
			aP0[i] = m_adInitPr[m_oMetaModel.m_aiClassStateIndex[iClass][i]] / tmp;
	
	m_adScale[0]=0.0;
	for( i=0; i<m; i++ ) { 
		m_adAlpha[0][i] = aP0[i];
		m_adScale[0] += m_adAlpha[0][i];
		}
	m_adScale[0]=1.0/ m_adScale[0];
	for (i=0; i<m; i++) 
		m_adAlpha[0][i] *= m_adScale[0];

	//----------------
	CMilModelData * pData;
	double dLL = log(m_adScale[0]);
	
	for( k=0; k<iDwells && ! m_bCancelled; k++) {
		k2 = 2*k;
		
		pData = &(m_pData[ aiStimIndex==NULL ? 0 : aiStimIndex[k] ]);
		
		//-- Mil paper eq.(5)	G(a.k,a.k+1) a.k.a. Gab
		//-- todo - consider precalcing for tDwells which appear >1 time.
		Array2D<double> aTemp( m, m, 0.0 );
		mil_expQ( m, 
				pData->m_aW[iClass], 
				pData->m_adP[iClass], 
				pData->m_adP1[iClass], 
				dSampling*aiDuration[k] - m_pOptionsMil->m_dDeadTime, aTemp );

		//-- Mil paper eq.(3)  alpha(k)=alpha(k-1) * G(a.k,a.k+1)
		//-- alpha[2*k+1] = ( alpha[2*k] * aTemp ) scaled by maximum element
		matmult( m, m, (double *)m_adAlpha[k2], (double **)aTemp, (double *)(m_adAlpha[k2+1]) );
		dScale = (1.0/max(1.0e-10, Sum(0,m-1,m_adAlpha[k2+1])));
		dLL += log( m_adScale[k2+1]=dScale );
		for( j=0; j<m; j++ ) 
			m_adAlpha[k2+1][j] *= dScale;
		
		//-- 2nd part of eq.(5) - * Qab
		//-- Create subarray of rows/cols from m_aQQe		
		if( k<iDwells-1 ) {
			iNextClass = aiDwell[k+1];
			n  = m_oMetaModel.m_aiClassStateCount[iNextClass];
			}
		else {
			n = m_iMetaStates - m_oMetaModel.m_aiClassStateCount[iClass];
			}
		//-- alpha[2*k+2] = ( alpha[2*k+1] * a ) scaled by maximum element
		matextractmult( m, 
					n, 
					(double *)m_adAlpha[k2+1], 
					m_oMetaModel.m_aiClassStateIndex[iClass], 
					(m_oMetaModel.m_aiClassStateIndex[iNextClass])+( k==iDwells-1?m_iMetaStates:0),
					(double **)pData->m_aQQe, (double *)(m_adAlpha[k2+2]) );
		dScale = (1.0/max(1.0e-6, Sum(0,n-1,m_adAlpha[k2+2])));
		dLL += log( m_adScale[k2+2]=dScale );
		for( j=0; j<n; j++ ) 
			m_adAlpha[k2+2][j] *= dScale;
		
		iClass = iNextClass;
		m  = n;
		}

	//-- Output the alpha values to a file for debugging purposes
	#ifdef _DEBUG
	FILE * h = fopen( "c:\\AlphaG.txt", "w" );
	fprintf( h, "Dwells=%d DeadTime=%f\n", iDwells, m_pOptionsMil->m_dDeadTime );
	for( k=0; k<iDwells && k<100; k++ ) {
		fprintf( h, "dwell=%d class=%d [%d] t=%10.5f Scale=%f,%f ", k, int(aiDwell[k]),
			m_oMetaModel.m_aiClassStateCount[int(aiDwell[k])], 
			double(dSampling*aiDuration[k]- m_pOptionsMil->m_dDeadTime), m_adScale[2*k], m_adScale[2*k+1] );
		
		fprintf( h, "{ " );
		for( i=0; i<m_iMetaStates; ++i ) 
			fprintf( h, "%f ", m_adAlpha[2*k][i] );
		fprintf( h, "} { " );
		for( i=0; i<m_iMetaStates; ++i ) 
			fprintf( h, "%f ", m_adAlpha[2*k+1][i] );
		fprintf( h, "}\n" );
		}
	fclose(h);	
	#endif	
	

	if( ! m_bCancelled )
		return m_pOptionsMil->m_dLL = dLL;
	return 0.0;
	}
	
void CTaskMil :: EvaluateGradient( double * adGradient ) { 
	int iDwells=m_pOptionsMil->m_oEventList.m_iEvents;
	short * aiDwell=m_pOptionsMil->m_oEventList.m_aiClass;
	int * aiDuration=m_pOptionsMil->m_oEventList.m_aiDuration;
	double dSampling=m_pOptionsMil->m_dSampling;

	mil_backward( iDwells, aiDwell, aiDuration, dSampling );

	//-- For each stimulus set (at least one ) - calc derivatives
	int iStims = m_pOptionsMil->m_oEventList.m_iStimCount;

	//-- back fill the derivatives to qqe, then qe, then q, then x, then z
	mil_dlogLL_qe( );
	
	for( int j=0; j<m_iZSize; j++ )
		adGradient[j] = 0.0;
		
	for( int iSet=0; iSet==0 || iSet<iStims; ++iSet ) {
		CMilModelData * pData = &m_pData[iSet];
			
		for( int k=0; k<m_iMetaPaths; k++ )
			pData->m_df_qq[k] = SumProduct( (int)m_iMetaStates, (int)m_iMetaStates, (double **)pData->m_df_qqe, (double **)pData->m_dQQe_qq[k] );

		pData->m_df_q = matmult( pData->m_df_qq, pData->m_dqq_q );

		pData->m_df_x = matmult( pData->m_df_q, pData->m_dq_x );

		pData->m_df_z = matmult( pData->m_df_x, pData->m_dx_z );

		for( int j=0; j<m_iZSize; j++ ) {
			adGradient[j] -= pData->m_df_z[j];
			/*
			adGradient[j] -= ( pData->m_dMagnitude * pData->m_df_z[j] );
			*/
			/*
			#ifdef _DEBUG
			char buf[100];
			sprintf( buf, "Set=%d Gradient[%d]=%f\n", iSet, j, adGradient[j] );
			OutputDebugString( buf );
			#endif
			*/
			}

		/*
		#ifdef _DEBUG
		char buf[100];
		sprintf( buf, "=============================Metapaths=%d\n", m_iMetaPaths );
		OutputDebugString( buf );
		OutputDebugArray( pData->m_aW, "W" );
		OutputDebugArray( pData->m_adP, "P" );
		OutputDebugArray( pData->m_adP1, "P1" );
		OutputDebugArray( pData->m_dQQe_qq, "dQQe_qq");
		OutputDebugArray( pData->m_df_qqe, "df_qqe" );
		OutputDebugVector( pData->m_df_qq, "df_qq" );
		OutputDebugVector( pData->m_df_q, "df_q" );
		OutputDebugVector( pData->m_df_x, "df_x" );
		OutputDebugVector( pData->m_df_z, "df_z" );
		#endif
		*/
		}
		
	/*
	//-- todo  Averaging gradients from all stim models.  Is this the correct thing to do ? 
	if( m_dMagnitudeSum!=0.0 ) 
		for( int j=0; j<m_iZSize; j++ ) 
			adGradient[j] /= m_dMagnitudeSum ;
	*/

	}

bool CTaskMil :: ZtoX( CMilModelData * pData, double * z, bool bUpdateRates ) { 
	//----- ZtoX
	veccopy( m_iXSize, (double *)pData->m_Vct, (double *)pData->m_X );
	matmult( m_iXSize, m_iZSize, (double **)pData->m_Mtx, (double *)z, (double *)pData->m_X, true );

	//TODO - remove this, then calculate not from model 
	// fix to use stimulus undo for filling rates 
	if( bUpdateRates ) { 
		for( int iP=0; iP<m_iOrigRates; ++iP ) {
			CModelRate * pRate = m_pOptionsMil->m_oModel.Rate( iP );
			pRate->m_adK0[0] = exp(pData->m_X[iP*4+0]);
			pRate->m_adK1[0] = pData->m_X[iP*4+1];
			pRate->m_adK0[1] = exp(pData->m_X[iP*4+2]);
			pRate->m_adK1[1] = pData->m_X[iP*4+3];
			}
		}
	/*TODO 
	bool bResult=true;
	if( fabs(xmax[i]-xmin[i])>1.0e-5 && (x[i]<xmin[i] || x[i]>xmax[i]) )
		bResult=false;
	return bResult; 
	*/
	return true; 
	}

//-----------------------------------------------------------------
// For DFP minimization, MIL evaluation and gradient calculation
/* 
Note: ZtoX updates X[] values.  MilEvalFunc only is used in dfpmin
which is only used for non-stimulus dependent MIL.  To use dfpmin
with multiple stimulii levels and analytical derivatives, need to 
combine gradients and ZtoX on a set of models.... 
*/
double MilEvalFunc( double * z, void * pTask  ) { 
	//-- process model z from dfp optimizer 
	CTaskMil * pMil = ((CTaskMil *)pTask);
	COptionsMil * pOpt = pMil->m_pOptionsMil;

	//-- For each stimulus set (at least one) - redo model calculate
	// m_iStimCount should not be > 1 yet, cannot handle it in derivatives.
	for( int iSet=0; iSet==0 || iSet<pOpt->m_oEventList.m_iStimCount; ++iSet ) {
		pMil->ZtoX( &pMil->m_pData[iSet], z, iSet==0 );
		pMil->ModelCalculate( &pOpt->m_oModel,  pOpt->m_dDeadTime, &pMil->m_pData[iSet] );
		}
		
	return pMil->EvaluateLL( );
	}
	
void MilEvalGrad( double * z, double * adGradient, void * pTask  ) { 
	CTaskMil * pMil = ((CTaskMil *)pTask);

	pMil->EvaluateGradient( adGradient );

	char buf[1024];
	sprintf( buf, "MilEvalGrad : iterations=%d/%d LL=%f Gradient={%10.7f,%10.7f,%10.7f,%10.7f,...}\n", 
		pMil->m_pOptionsMil->m_iIterations, 
		pMil->m_pOptionsMil->m_iMaxIter, 
		pMil->m_pOptionsMil->m_dLL, 
		adGradient[0], adGradient[1], adGradient[2], adGradient[3] );
	AddMessage( REPTYPE_RESULT, buf );
	}


void CTaskMil :: OptimizeAnalytical(  ) { 
	// TODO - where is pMil->m_dConvGrad checked ? 
	//-- Update the pct complete 
	// PostMessage( pTask->m_hDlg, WM_TASKSETPCT, MulDiv(iSeg,100,iSegs), 0 );
	//-- Initiate a DFP minimization 
	int iRet = dfpmin( 
		(double *) m_pData[0].m_Z,					// starting pt. (Z) / result
		m_iZSize,						// size of p (iZSize)
		m_pOptionsMil->m_dConvLL,		// convergence requirement for gradient
		m_pOptionsMil->m_iMaxIter,		// max iterations
		&(m_pOptionsMil->m_dLL ), 		// minimum value achieved
		MilEvalFunc,					// function to minimize
		MilEvalGrad,					// gradient calculation				
		(void *)this,					// data to pass to lnsrch (CMil *)
		&(m_pOptionsMil->m_iIterations),// iterations
		&m_bCancelled					// user cancel flag 
		);

	switch( iRet ) {
		case DFPMIN_CANC :
			break;
		case DFPMIN_EXIT :
			break;
		case DFPMIN_ITER :
			break;
		case DFPMIN_GTOL :
			break;
		}
	}
	
//-----------------------------------------------------------------
// Todo - modelCalculate for all stim models
double CTaskMil :: EvaluateStep( int iRate, int iDir, int iK, double dStep ) { 
	CModelRate * pRate = m_pOptionsMil->m_oModel.Rate(iRate);
	double * pData = (iK==0) ? &(pRate->m_adK0[iDir]) : &(pRate->m_adK1[iDir]);
	double dPrev = *pData;
	*pData += dStep;
ModelMultipleSetup( m_pOptionsMil );
//	ModelCalculate( &m_pOptionsMil->m_oModel, m_pOptionsMil->m_dDeadTime, &m_pData[0] );
	EvaluateLL( );
	*pData = dPrev;
	return m_pOptionsMil->m_dLL;
	}
	
double CTaskMil :: EvaluateSteps( double * aStep, int iRates, double dInitLL ) { 
	double dPrevLL = dInitLL, dLL;
	int iResult=0;
	for( int iMul=1; iMul<1000; iMul*=2 ) { 
		//-- increment the rates 
		for( int iRate=0; iRate<iRates; ++iRate ) { 
			CModelRate * pRate = m_pOptionsMil->m_oModel.Rate(iRate);
			for( int iDir=0; iDir<2; ++iDir ) { 
				pRate->m_adK0[iDir] += aStep[iRate*4+iDir*2] * iMul;
				pRate->m_adK1[iDir] += aStep[iRate*4+iDir*2+1] * iMul;
				}
			}
				
		//-- calc new LL 
		ModelCalculate( &m_pOptionsMil->m_oModel, m_pOptionsMil->m_dDeadTime, &m_pData[0] );
		dLL = EvaluateLL( );

		//-- too far - roll back the rates
		if( dLL > dPrevLL ) { 
			for( int iRate=0; iRate<iRates; ++iRate ) { 
				CModelRate * pRate = m_pOptionsMil->m_oModel.Rate(iRate);
				for( int iDir=0; iDir<2; ++iDir ) { 
					pRate->m_adK0[iDir] -= aStep[iRate*4+iDir*2] * iMul;
					pRate->m_adK1[iDir] -= aStep[iRate*4+iDir*2+1] * iMul;
					}
				}
			break;
			}
			
		dPrevLL = dLL;
		iResult=iMul;
		}
		
	return dPrevLL;
	}
			
void CTaskMil :: OptimizeNumerical( ) { 
	const double LL_TOL = 1.0;
	double dLLPos, dLLNeg, dLLNew;

	ModelMultipleSetup( m_pOptionsMil );
	EvaluateSetup( m_pOptionsMil->m_oEventList.m_iEvents );
	EvaluateLL( );
	
	double dLL = m_pOptionsMil->m_dLL; // EvaluateStep( 0, 0, 0, 0.0 );
ODBG("Init", m_pOptionsMil );

	double dStep[2]={m_pOptionsMil->m_dK0Step,m_pOptionsMil->m_dK1Step};
	double dSum;
	int i, iIndex, iRates = m_pOptionsMil->m_oModel.m_aRates.Count();
	
//	double * aStep = new double[iRates*4];
//	double * aDelta = new double[iRates*4];
	
	m_pOptionsMil->m_oModel.SetConstraintStatus();
	
	m_pOptionsMil->m_iIterations=0;
	dLLNew=dLL;
	do { 
		dLL = dLLNew;
		
		int iErr=0;
		
		//-- determine dLL/dX for all X ( k0 and k1 )
		for( int iRate=0; iRate<iRates; ++iRate ) { 
			CModelRate * pRate = m_pOptionsMil->m_oModel.Rate(iRate);
			for( int iDir=0; iDir<2; ++iDir ) { 
				for( int iK=0; iK<2; ++iK ) {		// Which component of rate - k0 or k1
					iIndex = iRate*4+iDir*2+iK;
					if( (iK==0?pRate->m_abConstrainedK0:pRate->m_abConstrainedK1)[iDir] ) { 
//						aStep[iIndex]=aDelta[iIndex]=0.0;
						iErr++;
						}
					else { 
						//-- determine direction of improving likelihood for change to this component
						dLLPos = EvaluateStep( iRate, iDir, iK, dStep[iK] );
						dLLNeg = EvaluateStep( iRate, iDir, iK, -dStep[iK] );
						double * pData = (iK==0) ? &(pRate->m_adK0[iDir]) : &(pRate->m_adK1[iDir]);

						if( dLLNew-dLLPos<=1.0e-6 && dLLNew-dLLNeg<=1.0e-6 ) {	// negative or nearly 0 
							//-- neither direction improves likelihood
//							aStep[iIndex]=aDelta[iIndex]=0.0;
							iErr++;
							}
						else if( dLLNew-dLLPos > dLLNew-dLLNeg ) { 
							//-- positive step is better
//							aStep[iIndex]=dStep[iK];
//							aDelta[iIndex]=dLL-dLLPos;
*pData += dStep[iK];
dLLNew=dLLPos;
							}
						else { 
							//-- negative step is better
//							aStep[iIndex]=-dStep[iK];
//							aDelta[iIndex]=dLL-dLLNeg;
*pData -= dStep[iK];
dLLNew=dLLNeg;
							}
ModelMultipleSetup( m_pOptionsMil );
EvaluateLL( ); 
ODBG("Midl", m_pOptionsMil );
						}
					}
				}
			}
ODBG("Finl", m_pOptionsMil );

		if( iErr==iRates*4 ) { 
			#ifdef _DEBUG
			char bufErr[100];
			sprintf( bufErr, "Step too large : [%f,%f] \n", dStep[0], dStep[1] );
			OutputDebugString( bufErr );
			#endif
			dStep[0] *= 0.5;
			dStep[1] *= 0.5;
			}
		else { 
			//-- determine initial combined step 
//			dSum=0.0001;
//			for( i=0; i<iRates*4; ++i ) 
//				dSum += aDelta[i];
//			for( i=0; i<iRates*4; ++i ) 
//				aStep[i] *= aDelta[i]/dSum;
			
			//-- Go in that direction as far as reasonable
//			dLLNew = EvaluateSteps( aStep, iRates, dLL );
//			if( dLLNew>=dLL ) {
//				dStep[0] *= 0.5;
//				dStep[1] *= 0.5;
//				}
//			m_pOptionsMil->m_oModel.ApplyConstraints();

//			#ifdef _DEBUG
//			//-- show progress		
//			char buf[1000];
//			sprintf( buf, "Iter=%d Step=%f,%f LL=%f LLNew=%f X={%f,%f,%f,%f} Delta={%f,%f,%f,%f}\n", 
//						m_pOptionsMil->m_iIterations, 
//						dStep[0], dStep[1], dLL, dLLNew, 
//						m_pOptionsMil->m_oModel.Rate(0)->m_adK0[0], 
//						m_pOptionsMil->m_oModel.Rate(0)->m_adK0[1], 
//						m_pOptionsMil->m_oModel.Rate(1)->m_adK0[0], 
//						m_pOptionsMil->m_oModel.Rate(1)->m_adK0[1], 
//						aStep[0], aStep[2], aStep[4], aStep[6]);
//			OutputDebugString( buf );
//			#endif
			}
		} while( ! m_bCancelled
				&& ++m_pOptionsMil->m_iIterations<m_pOptionsMil->m_iMaxIter
				&& ( dLL-dLLNew>LL_TOL || dStep[0]>=m_pOptionsMil->m_dK0Step/256.0 ) );
				
//	delete [] aStep;
//	delete [] aDelta;
ModelMultipleSetup( m_pOptionsMil );
EvaluateLL( );
ODBG("Exit", m_pOptionsMil );
	}

//-----------------------------------------------------------------
DWORD WINAPI TaskMil( LPVOID pParam ) { 
	//-----
	CTaskMil * pTask = (CTaskMil *) pParam;

	COptionsMil * pOpt = pTask->m_pOptionsMil;
	pOpt->m_iChannels = max( pOpt->m_iChannels, 1 );
	
	pTask->m_bCancelled=false;

	pTask->m_bAnalyticalDerivative = ! pOpt->m_bNumericDerivatives;

	//-- Allocate ModelData for number of Stimulus sets 
	pTask->m_pData = new CMilModelData[ max(1,pOpt->m_oEventList.m_iStimCount) ];
	
	//-- create alpha / beta (  & Scale ) arrays 	
	if( pOpt->m_iMode == OPT_MODE_EVALUATE ) {
		pTask->ModelMultipleSetup( pOpt );
		pTask->EvaluateSetup( pOpt->m_oEventList.m_iEvents );
		pTask->EvaluateLL( );

		char buf[100];
		sprintf( buf, "LL=%f\n", pOpt->m_dLL );
		AddMessage( REPTYPE_RESULT, buf );
		}
	
	else if( pOpt->m_iMode == OPT_MODE_OPTIMIZE ) { 
		pOpt->m_dLL=0.0;
		pOpt->m_iIterations=0;
		if(	! pTask->m_bAnalyticalDerivative /*pOpt->m_oEventList.m_iStimCount > 1*/ )
			pTask->OptimizeNumerical( );		// -- use numerical derivatives
			
		else {
			pTask->ModelMultipleSetup( pOpt );
			pTask->EvaluateSetup( pOpt->m_oEventList.m_iEvents );
			pTask->OptimizeAnalytical( );		// -- use analytical derivatives
			}
		
		CModel * pOrig = pOpt->m_pOriginalModel;
		for( int iRate=0; iRate<pOrig->m_aRates.Count(); ++iRate ) {
			CModelRate * pOrigRate = pOrig->Rate(iRate);
			CModelRate * pRate = pOpt->m_oModel.Rate(iRate);
			#ifdef _DEBUG
			char buf[256];
			sprintf( buf, "%d : k0(0)=%f->%f  k0(1)=%f->%f  k1(0)=%f->%f  k1(1)=%f->%f \n", 
					iRate, 
					pOrigRate->m_adK0[0],pRate->m_adK0[0],
					pOrigRate->m_adK0[1],pRate->m_adK0[1],
					pOrigRate->m_adK1[0],pRate->m_adK1[0],
					pOrigRate->m_adK1[1],pRate->m_adK1[1]);
			OutputDebugString( buf );
			#endif			
			pOrigRate->m_adK0[0]=pRate->m_adK0[0];
			pOrigRate->m_adK1[0]=pRate->m_adK1[0];
			pOrigRate->m_adK0[1]=pRate->m_adK0[1];
			pOrigRate->m_adK1[1]=pRate->m_adK1[1];
			}
			
		PostMessage( pTask->m_hModelWnd, WM_MODEL_CHANGED, 1, 0 );		// wparam=1 if window should be activated.
		char buf[100];
		sprintf( buf, "Optimize done : iterations=%d/%d LL=%f\n", pOpt->m_iIterations, pOpt->m_iMaxIter, pOpt->m_dLL );
		AddMessage( REPTYPE_RESULT, buf );
		}
		
	//-- 
	PostMessage( (HWND)pTask->m_hDlg, WM_TASKCOMPLETE, pTask->m_bCancelled?1:0, 0 );
	
	return -1;
	}	
